#include "subdivision.h"

Subdivision::Subdivision(QOpenGLFunctions_4_3_Core *f)
{
    this->f =f;
}

Subdivision::~Subdivision()
{
    delete shader;
}

void Subdivision::init() {
    shader = initComputeShaderProgram();
}

QOpenGLShaderProgram *Subdivision::initComputeShaderProgram(){
    QString source = QLatin1String(":/subdivision.compute");

    qDebug()<<"Compiling compute shader ...";
    QOpenGLShader *computeShader = new QOpenGLShader(QOpenGLShader::Compute);
    if(!computeShader->compileSourceFile(source)){
        qCritical()<<"Compute Shader"<<source<<"failed"<<computeShader->log();
        exit(5);
    }

    qDebug()<<"Adding compute shader ...";
    QOpenGLShaderProgram *shader = new QOpenGLShaderProgram();
    shader->addShader(computeShader);
    qDebug()<<"Linking compute shader ...";
    if(!shader->link()){
        qCritical()<<"Linking compute shader failed:"<<shader->log();
        exit(5);
    }
    qDebug()<<"Linking compute shader done";

    return shader;
}

Mesh *Subdivision::subdivide(Mesh *mesh) {
    precomputeTables(mesh);
    runShader(mesh);
    return NULL;
}

void Subdivision::precomputeTables(Mesh *mesh) {
    Mesh::Node root = mesh->getRootNode();

    int firstMeshEntryIndex = -1;
    bool meshExists = root.getFirstMeshIndex(firstMeshEntryIndex);
    if (!meshExists){
        qDebug()<<"No Mesh found. Abort subdiv table precomputation.";
        return;
    }
    qDebug()<<"Mesh found. Index:"<<firstMeshEntryIndex;

    //die drei zeilen machen das bild schwarz. wenn man davor returnt tuts.
    Mesh::MeshEntry firstMeshEntry = mesh->getMeshEntry(firstMeshEntryIndex);
    QVector<unsigned int> ib = firstMeshEntry.getIndexBuffer();
    qDebug()<<"Index Buffer: "<<ib;

    return;
    //todo find adjacent triangles:
/*
    for (int i = 0; i < ib.length(); i+=3){
        //schaue alle dreiecke an

        unsigned int x, y, z;
        x_index = ib[i];
        y_index = ib[i+1];
        z_index = ib[i+2];

        //get vertices x,y,z from vertex buffer

        for (int j = i+3; j < ib.length(); j+= 3){
            //search all following triangles for common edge

            a_index = ib[j];
            b_index = ib[j+1];
            c_index = in[j+2];

            //get vertices for a,b,c from vertex buffer

            //for xy

            //for yz

            //for zx
        }

    }
     *
     * */

}

void Subdivision::runShader(Mesh *mesh) {
    /*
     * This shader will set the output buffer to its global ID times hundred + the local id + offset.
     *
     * So, for offset = 0 we will get 0, 101, 200, 301 and so on as there's two threads in a local group.
     */

    qDebug()<<"Running compute shader";
    int num = 6 * 2; // must be a multiple of two as we do not handle the other case
    GLfloat offset = 3.0f;

    shader->bind();

    // Create an input buffer and set it to the value of offset
    GLuint offsetBufferID;
    f->glGenBuffers(1, &offsetBufferID);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, offsetBufferID);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, sizeof(GLfloat), &offset, GL_DYNAMIC_DRAW);

    // Create an output buffer
    GLuint bufferID;
    f->glGenBuffers(1, &bufferID);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, bufferID);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, sizeof(GLfloat) * num, NULL, GL_DYNAMIC_DRAW);

    // Run the shader
    f->glDispatchCompute(num/2, 1, 1);

    // Wait for the shader to complete and the data to be written back to the global memory
    f->glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

    // Unbind the buffers from the shader
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);

    // Map the output buffer so we can read the results
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, bufferID);
    GLfloat *ptr;
    ptr = (GLfloat *) f->glMapBuffer(GL_SHADER_STORAGE_BUFFER, GL_WRITE_ONLY);
    for (int i = 0; i < num; i++) {
        qDebug()<<i<<ptr[i];
    }
    f->glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);

    // Delete the buffers the free the resources
    f->glDeleteBuffers(1, &offsetBufferID);
    f->glDeleteBuffers(1, &bufferID);

    shader->release();
}
